import openai
import openai_async
import os
import sys
import textwrap

import chatdbg_utils

async def why(self, arg):
    user_prompt = "Explain what the root cause of this error is, given the following source code and traceback, and generate code that fixes the error."
    user_prompt += "\n"
    user_prompt += "source code:\n```\n"
    stack_trace = ""
    stack_frames = len(self.stack)
    try:
        import sys

        exception_name = sys.exc_info()[0].__name__
        exception_value = sys.exc_info()[1]
    except:
        print(
            "The command 'why' only works when there is an uncaught exception. Try running 'python3 -m chatdbg -c continue'."
        )
        return
    for frame_lineno in self.stack:
        import inspect

        frame, lineno = frame_lineno
        # Only include frames for files in the same directory as the program being debugged.
        # TODO: add a --program-path option as in Scalene
        if not frame.f_code.co_filename.startswith(os.path.dirname(sys.argv[0])):
            stack_frames -= 1
            continue
        try:
            # user_prompt += '#' + '-' * 60 + '\n'
            lines = inspect.getsourcelines(frame)[0]
            for index, line in enumerate(lines, frame.f_code.co_firstlineno):
                user_prompt += "  "
                user_prompt += line.rstrip() + "\n"
                if index == lineno:
                    leading_spaces = len(line) - len(line.lstrip())
                    stack_trace += f"{stack_frames}: " + line.strip() + "\n"
                    # Degrade gracefully when using older Python versions that don't have column info.
                    try:
                        positions = inspect.getframeinfo(frame).positions
                    except:
                        positions = None
                    if positions:
                        stack_trace += (
                            " " * len(str(stack_frames))
                            + "  "
                            + " " * (positions.col_offset - leading_spaces)
                            + "^" * (positions.end_col_offset - positions.col_offset)
                            + "\n"
                    )
                if index >= lineno:
                    break
        except:
            pass
        stack_frames -= 1
    user_prompt += "```\n"
    user_prompt += "stack trace:\n"
    user_prompt += f"```\n{stack_trace}```\n"
    user_prompt += f"Exception: {exception_name} ({exception_value})\n"

    #print(user_prompt)
    #return
    
    import httpx

    model = chatdbg_utils.get_model()
    if not model:
        return
    
    text = ""
    try:
        completion = await openai_async.chat_complete(
            openai.api_key or "",
            timeout=30,
            payload={
                "model": model,
                "messages": [{"role": "user", "content": user_prompt}],
            },
        )
        json_payload = completion.json()
        if not 'choices' in json_payload:
            raise openai.error.AuthenticationError
        text = json_payload["choices"][0]["message"]["content"]
    except (openai.error.AuthenticationError, httpx.LocalProtocolError):
        print(
            "You need a valid OpenAI key to use commentator. You can get a key here: https://openai.com/api/"
        )
        print("Set the environment variable OPENAI_API_KEY to your key value.")
    except Exception as e:
        print(f"EXCEPTION {e}")
        pass
    print(chatdbg_utils.word_wrap_except_code_blocks(text))



import os
import struct


def is_debuggable_binary(filepath):
    # Check if the file exists
    if not os.path.exists(filepath):
        print(f"File '{filepath}' does not exist.")
        return False

    # Read the first 4 bytes of the file to check for the Mach-O magic number
    with open(filepath, 'rb') as file:
        magic_number = file.read(4)
    
    if magic_number != b'\xFE\xED\xFA\xCE' and magic_number != b'\xCE\xFA\xED\xFE':
        print(f"File '{filepath}' is not a Mach-O binary.")
        return False

    # Extract the number of load commands
    with open(filepath, 'rb') as file:
        file.seek(16)
        num_load_commands = struct.unpack('I', file.read(4))[0]

    # Read the load commands to find the LC_DYSYMTAB
    with open(filepath, 'rb') as file:
        file.seek(32)
        for _ in range(num_load_commands):
            cmd = struct.unpack('I', file.read(4))[0]
            cmdsize = struct.unpack('I', file.read(4))[0]

            if cmd == 0xb:  # LC_DYSYMTAB
                file.seek(12, 1)  # Skip the indirect symbol table offset, number of indirect symbol table entries, and other fields
                flags = struct.unpack('I', file.read(4))[0]

                if flags & 0x2:  # Check if the flag for stripped symbols is not set
                    return True  # Debug info section found
                else:
                    return False  # Debug info section not found

    return False  # Debug info section not found


# Usage example
executable_path = 'a.out' # /path/to/executable'  # Replace with the actual path to the executable
is_debuggable = is_debuggable_binary(executable_path)
print(f"Is the executable debuggable? {is_debuggable}")



#!env python3
import lldb
import asyncio

# xcrun python3 -m pip install whatever


def __lldb_init_module(debugger, internal_dict):
    debugger.HandleCommand("command script add -f chatdbg_lldb.why why")


# From lldbinit
def get_process():
    """
    A read only property that returns an lldb object
    that represents the process (lldb.SBProcess)that this target owns.
    """
    return get_target().process


def get_frame():
    frame = None
    # SBProcess supports thread iteration -> SBThread
    for thread in get_process():
        if (thread.GetStopReason() != lldb.eStopReasonNone) and (
            thread.GetStopReason() != lldb.eStopReasonInvalid
        ):
            frame = thread.GetFrameAtIndex(0)
            break
    # this will generate a false positive when we start the target the first time because there's no context yet.
    if not frame:
        # print("[-] warning: get_frame() failed. Is the target binary started?")
        pass
    return frame


def get_thread():
    thread = None
    # SBProcess supports thread iteration -> SBThread
    for _thread in get_process():
        if (_thread.GetStopReason() != lldb.eStopReasonNone) and (
            _thread.GetStopReason() != lldb.eStopReasonInvalid
        ):
            thread = _thread

    if not thread:
        # print("[-] warning: get_thread() failed. Is the target binary started?")
        pass

    return thread


def get_target():
    target = lldb.debugger.GetSelectedTarget()
    if not target:
        # print("[-] error: no target available. please add a target to lldb.")
        return None
    return target


#
def read_lines(file_path, start_line, end_line):
    with open(file_path, "r") as f:
        lines = f.readlines()
        lines = [line.rstrip() for line in lines]  # remove trailing newline characters

    start_line = max(0, start_line - 1)  # convert to 0-based indexing
    end_line = min(len(lines), end_line)  # ensure end_line is within range

    return "\n".join(lines[start_line:end_line])


import re


def get_function_by_line(source_code, line_number):
    """
    Given a source code and a line number, this function returns the function that contains the line number.
    """
    # Remove all single and multi-line comments
    source_code = re.sub(r"\/\/.*", "", source_code)
    source_code = re.sub(r"\/\*[\s\S]*?\*\/", "", source_code)

    # Remove all whitespaces and newlines
    source_code = re.sub(r"\s+", "", source_code)

    # Find all function definitions
    function_regex = r"(?P<signature>(static|const)?[\w\s\*]+)(?P<name>\w+)\((?P<args>[\w\s\*,]*)\)(?P<body>{[\s\S]*?})"
    functions = re.finditer(function_regex, source_code)

    # Find the function that contains the line number
    for function in functions:
        function_body = function.group("body")
        line_count = function_body.count("\n")

        print("checking ", function_body)
        if line_number <= line_count:
            return (
                function.group("signature")
                + function.group("name")
                + "("
                + function.group("args")
                + ")"
                + function_body
            )

        line_number -= line_count

    return None


def stackTrace(debugger):
    import os

    target = get_target()
    if not target:
        return ""
    thread = get_thread()
    if not thread:
        # Not running
        return ""

    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        return ""

    frame = thread.GetFrameAtIndex(0)
    stack_trace = ""
    source_code = ""
    for index, frame in enumerate(thread):
        function = frame.GetFunction()
        if not function:
            # No debug info
            continue

        # Check for debugging info https://lldb.llvm.org/python_api/lldb.SBFunction.html
        func_name = frame.GetFunctionName()
        line_entry = frame.GetLineEntry()
        file_spec = line_entry.GetFileSpec()
        file_name = file_spec.GetFilename()
        directory = file_spec.GetDirectory()
        full_file_name = os.path.join(directory, file_name)
        line_num = line_entry.GetLine()
        col_num = line_entry.GetColumn()
        stack_trace += (
            f"frame {index}: {func_name} at {file_name}:{line_num}:{col_num}\n"
        )
        source_code += read_lines(full_file_name, line_num - 10, line_num) + "\n"
        source_code += "-" * (col_num - 1) + "^" + "\n\n"

    error_reason = thread.GetStopDescription(255)
    return (source_code, stack_trace, error_reason)


import sys

import openai
import openai_async
import os
import sys
import textwrap


def word_wrap_except_code_blocks(text: str) -> str:
    """Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split("\n\n")
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # If this paragraph starts and ends with a code block, add it as is.
        if paragraph.startswith("```") and paragraph.endswith("```"):
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the beginning of a code block add it as is.
        if paragraph.startswith("```"):
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the end of a code block stop skipping text.
        if paragraph.endswith("```"):
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue
        # If we are currently in a code block add the paragraph as is.
        if in_code_block:
            wrapped_paragraphs.append(paragraph)
        else:
            # Otherwise, apply text wrapping to the paragraph.
            wrapped_paragraph = textwrap.fill(paragraph)
            wrapped_paragraphs.append(wrapped_paragraph)
    # Join all paragraphs into a single string
    wrapped_text = "\n\n".join(wrapped_paragraphs)
    return wrapped_text


async def explain(source_code, traceback, exception):
    import httpx

    user_prompt = "Explain what the root cause of this error is, given the following source code and traceback, and propose a fix."
    user_prompt += "\n"
    user_prompt += "source code:\n```\n"
    user_prompt += source_code + "\n```\n"

    user_prompt += traceback + "\n\n"

    user_prompt += "stop reason = " + exception + "\n"

    text = ""
    try:
        completion = await openai_async.chat_complete(
            openai.api_key,
            timeout=30,
            payload={
                "model": "gpt-3.5-turbo",
                "messages": [{"role": "user", "content": user_prompt}],
            },
        )
        json_payload = completion.json()
        text = json_payload["choices"][0]["message"]["content"]
    except (openai.error.AuthenticationError, httpx.LocalProtocolError):
        print()
        print(
            "You need an OpenAI key to use commentator. You can get a key here: https://openai.com/api/"
        )
        print("Set the environment variable OPENAI_API_KEY to your key value.")
        import sys

        sys.exit(1)
    except Exception as e:
        print(f"EXCEPTION {e}")
        pass
    print(word_wrap_except_code_blocks(text))


def why(debugger, command, result, internal_dict):
    if not get_target():
        print("Must be attached to a program to ask `why`.")
        return
    thread = get_thread()
    if not thread:
        print("Must run the code first to ask `why`.")
        return
    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        print("Execution stopped at a breakpoint, not an error.")
        return

    the_trace = stackTrace(debugger)
    asyncio.run(explain(the_trace[0], the_trace[1], the_trace[2]))



#!env python3
import lldb
import asyncio
import os
import re
import sys

import pathlib
the_path = pathlib.Path(__file__).parent.resolve()

# The file produced by the panic handler if the Rust program is using the chatdbg crate.
rust_panic_log_filename = "panic_log.txt"

sys.path.append(os.path.abspath(the_path))

import chatdbg_utils

from typing import Tuple, Union

def __lldb_init_module(debugger: lldb.SBDebugger, internal_dict: dict) -> None:
    # Update the prompt.
    debugger.HandleCommand("settings set prompt '(ChatDBG lldb) '")

def is_debug_build(debugger, command, result, internal_dict) -> bool:
    """Returns False if not compiled with debug information."""
    target = debugger.GetSelectedTarget()
    if not target:
        return False

    has_debug_symbols = False
    for module in target.module_iter():
        for cu in module.compile_unit_iter():
            for line_entry in cu:
                if line_entry.GetLine() > 0:
                    has_debug_symbols = True
                    break
    return has_debug_symbols

def is_debug_build_prev(debugger, command, result, internal_dict) -> bool:
    target = debugger.GetSelectedTarget()
    if target:
        module = target.GetModuleAtIndex(0)
        if module:
            compile_unit = module.GetCompileUnitAtIndex(0)
            if compile_unit.IsValid():
                return True
    return False

   
# From lldbinit

def get_process() -> Union[None, lldb.SBProcess]:
    """
    Get the process that the current target owns.
    :return: An lldb object representing the process (lldb.SBProcess) that this target owns.
    """
    return get_target().process

def get_frame() -> lldb.SBFrame:
    """
    Get the current frame of the running process.

    :return: The current frame of the running process as an SBFrame object.
    """
    # Initialize frame variable to None
    frame = None
    for thread in get_process():
        # Loop through the threads in the process
        if thread.GetStopReason() != lldb.eStopReasonNone and thread.GetStopReason() != lldb.eStopReasonInvalid:
            # If the stop reason is not "none" or "invalid", get the frame at index 0 and break the loop.
            frame = thread.GetFrameAtIndex(0)
            break
    if not frame:
        # If frame is None, print a warning message
        # print("[-] warning: get_frame() failed. Is the target binary started?")
        # The warning message has been commented out, so just pass.
        pass
    # Return the current frame.
    return frame

def get_thread() -> lldb.SBThread:
    """
    Returns the currently stopped thread in the debugged process.
    :return: The currently stopped thread or None if no thread is stopped.
    """
    thread = None
    # Iterate over threads in the process
    for _thread in get_process():
        # Check if thread is stopped for a valid reason
        if _thread.GetStopReason() != lldb.eStopReasonNone and _thread.GetStopReason() != lldb.eStopReasonInvalid:
            thread = _thread
    if not thread:
        # No stopped thread was found
        pass
    return thread

def get_target() -> lldb.SBTarget:
    target = lldb.debugger.GetSelectedTarget()
    if not target:
        return None
    return target

def truncate_string(string, n):
    if len(string) <= n:
        return string
    else:
        return string[:n] + "..."

def buildPrompt(debugger: any) -> Tuple[str, str, str]:
    import os
    target = get_target()
    if not target:
        return ''
    thread = get_thread()
    if not thread:
        return ''
    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        return ''
    frame = thread.GetFrameAtIndex(0)
    stack_trace = ''
    source_code = ''
    
    # magic number - don't bother walking up more than this many frames.
    # This is just to prevent overwhelming OpenAI (or to cope with a stack overflow!).
    max_frames = 10

    index = 0
    for frame in thread:
        if index >= max_frames:
            break
        function = frame.GetFunction()
        if not function:
            continue
        full_func_name = frame.GetFunctionName()
        func_name = full_func_name.split('(')[0]
        arg_list = []
        type_list = []

        # Build up an array of argument values to the function, with type info.
        for i in range(len(frame.GetFunction().GetType().GetFunctionArgumentTypes())):
            arg = frame.FindVariable(frame.GetFunction().GetArgumentName(i))
            if not arg:
                continue
            arg_name = str(arg).split('=')[0].strip()
            arg_val = str(arg).split('=')[1].strip()
            arg_list.append(f"{arg_name} = {arg_val}")
            
        # Get the frame variables
        variables = frame.GetVariables(True, True, True, True)
        var_list = []
        
        for var in variables:
            name = var.GetName()
            value = var.GetValue()
            type = var.GetTypeName()
            # Check if the value is a pointer
            if var.GetType().IsPointerType():
                # Attempt to dereference the pointer
                try:
                    deref_value = var.Dereference().GetValue()
                    var_list.append(f"{type} {name} = {value} (*{name} = {deref_value})")
                except:
                    var_list.append(f"{type} {name} = {value}")
            
        line_entry = frame.GetLineEntry()
        file_spec = line_entry.GetFileSpec()
        file_name = file_spec.GetFilename()
        directory = file_spec.GetDirectory()
        full_file_name = os.path.join(directory, file_name)
        line_num = line_entry.GetLine()
        col_num = line_entry.GetColumn()

        max_line_length = 100
        
        try:
            lines = chatdbg_utils.read_lines(full_file_name, line_num - 10, line_num)
            stack_trace += truncate_string(f'frame {index}: {func_name}({",".join(arg_list)}) at {file_name}:{line_num}:{col_num}\n', max_line_length - 3) + '\n' # 3 accounts for ellipsis
            if len(var_list) > 0:
                stack_trace += "Local variables: " + truncate_string(','.join(var_list), max_line_length) + '\n'
            source_code += f'/* frame {index} in {file_name} */\n'
            source_code += lines + '\n'
            source_code += '-' * (chatdbg_utils.read_lines_width() + col_num - 1) + '^' + '\n\n'
        except:
            # Couldn't find the source for some reason. Skip the file.
            continue
        index += 1
    error_reason = thread.GetStopDescription(255)
    # If the Rust panic log exists, append it to the error reason.
    try:
        with open(rust_panic_log_filename, "r") as log:
            panic_log = log.read()
        error_reason = panic_log + "\n" + error_reason
    except:
        pass
    return (source_code, stack_trace, error_reason)

@lldb.command("why")
def why(debugger: lldb.SBDebugger, command: str, result: str, internal_dict: dict, really_run = True) -> None:
    """
    Root cause analysis for an error.
    """
    # Check if there is debug info.
    if not is_debug_build(debugger, command, result, internal_dict):
        print('Your program must be compiled with debug information (`-g`) to use `why`.')
        return
    # Check if program is attached to a debugger.
    if not get_target():
        print('Must be attached to a program to ask `why`.')
        return
    # Check if code has been run before executing the `why` command.
    thread = get_thread()
    if not thread:
        print('Must run the code first to ask `why`.')
        return
    # Check why code stopped running.
    if thread.GetStopReason() == lldb.eStopReasonBreakpoint:
        # Check if execution stopped at a breakpoint or an error.
        print('Execution stopped at a breakpoint, not an error.')
        return
    the_prompt = buildPrompt(debugger)
    asyncio.run(chatdbg_utils.explain(the_prompt[0], the_prompt[1], the_prompt[2], really_run))

@lldb.command("why_prompt")
def why_prompt(debugger: lldb.SBDebugger, command: str, result: str, internal_dict: dict) -> None:
    why(debugger, command, result, internal_dict, really_run=False)
    



import textwrap

def word_wrap_except_code_blocks(text: str) -> str:
    """
    Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split("\n\n")
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # Check for the presence of triple quotes in the paragraph
        if "```" in paragraph:
            # Split paragraph by triple quotes
            parts = paragraph.split("```")
            for i, part in enumerate(parts):
                # If we are inside a code block, do not wrap the text
                if in_code_block:
                    wrapped_paragraphs.append(part)
                else:
                    # Otherwise, apply text wrapping to the part
                    wrapped_paragraphs.append(textwrap.fill(part))
                # Toggle the in_code_block flag for each triple quote encountered
                if i < len(parts) - 1:
                    wrapped_paragraphs.append("```")
                    in_code_block = not in_code_block
        else:
            # If the paragraph does not contain triple quotes and is not inside a code block, wrap the text
            if not in_code_block:
                wrapped_paragraphs.append(textwrap.fill(paragraph))
            else:
                wrapped_paragraphs.append(paragraph)
    # Join all paragraphs into a single string
    wrapped_text = "\n\n".join(wrapped_paragraphs)
    return wrapped_text

# Example usage:
input_text = '''The root cause of this error is that the variable `x` is initialized
to 1.0, and then iteratively multiplied by `i` in the `for` loop in
the `fact` function. However, the loop starts from `i = 0.0`, which
means that `x` will always be multiplied by 0 in the first iteration,
resulting in `x` being 0.0. The subsequent `assert` statement checks
if `x` is not equal to 0.0, which will always fail.

To fix this error, we need to initialize `x` to 1.0 instead of 0.0.
This can be done by changing the initialization of `i` in the `for`
loop to start from 1.0 instead of 0.0.

Fixed source code: ``` /* frame 0 */ /* frame 1 */ /* frame 2 */ /*
frame 3 */ /* frame 4 */ #include <assert.h> #include <iostream>

float fact(float n) {   auto x = 1.0;   for (auto i = 1.0; i < n; i++)
{     x *= i;   }   assert(x != 0.0);   ^-----------------------------
------------------------------------------------ /* frame 5 */   for
(auto i = 1.0; i < n; i++) {     x *= i;   }   assert(x != 0.0);
return x; }


int main()
{
  std::cout << fact(100) << std::endl;
  ^-----------------------------------------------------------------------------
      ```

With this fix, `x` will correctly accumulate the product of `i`
values, starting from 1 instead of 0, and the `assert` statement will
not fail.'''

print(word_wrap_except_code_blocks(input_text))



def word_wrap_except_code_blocks(text: str) -> str:
    """Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split('\n\n')
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # If this paragraph starts and ends with a code block, add it as is.
        if paragraph.startswith('```') and paragraph.endswith('```'):
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the beginning of a code block add it as is.
        if paragraph.startswith('```'):
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the end of a code block stop skipping text.
        if paragraph.endswith('```'):
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue
        # If we are currently in a code block add the paragraph as is.
        if in_code_block:
            wrapped_paragraphs.append(paragraph)
        else:
            # Otherwise, apply text wrapping to the paragraph.
            wrapped_paragraph = textwrap.fill(paragraph)
            wrapped_paragraphs.append(wrapped_paragraph)
    # Join all paragraphs into a single string
    wrapped_text = '\n\n'.join(wrapped_paragraphs)
    return wrapped_text



import openai
import openai_async
import os
import sys
import textwrap

def get_model() -> str:
    all_models = ['gpt-4', 'gpt-3.5-turbo']
    
    if not 'OPENAI_API_MODEL' in os.environ:
        model = 'gpt-4'
    else:
        model = os.environ['OPENAI_API_MODEL']
        if model not in all_models:
            print(f'The environment variable OPENAI_API_MODEL is currently set to "{model}".')
            print(f'The only valid values are {all_models}.')
            return ""
        
    return model


def word_wrap_except_code_blocks(text: str) -> str:
    """
    Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split("\n\n")
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # Check for the presence of triple quotes in the paragraph
        if "```" in paragraph:
            # Split paragraph by triple quotes
            parts = paragraph.split("```")
            for i, part in enumerate(parts):
                # If we are inside a code block, do not wrap the text
                if in_code_block:
                    wrapped_paragraphs.append(part)
                else:
                    # Otherwise, apply text wrapping to the part
                    wrapped_paragraphs.append(textwrap.fill(part))
                # Toggle the in_code_block flag for each triple quote encountered
                if i < len(parts) - 1:
                    wrapped_paragraphs.append("```")
                    in_code_block = not in_code_block
        else:
            # If the paragraph does not contain triple quotes and is not inside a code block, wrap the text
            if not in_code_block:
                wrapped_paragraphs.append(textwrap.fill(paragraph))
            else:
                wrapped_paragraphs.append(paragraph)
    # Join all paragraphs into a single string
    wrapped_text = "\n\n".join(wrapped_paragraphs)
    return wrapped_text

def word_wrap_except_code_blocks_previous(text: str) -> str:
    """Wraps text except for code blocks.

    Splits the text into paragraphs and wraps each paragraph,
    except for paragraphs that are inside of code blocks denoted
    by ` ``` `. Returns the updated text.

    Args:
        text: The text to wrap.

    Returns:
        The wrapped text.
    """
    # Split text into paragraphs
    paragraphs = text.split('\n\n')
    wrapped_paragraphs = []
    # Check if currently in a code block.
    in_code_block = False
    # Loop through each paragraph and apply appropriate wrapping.
    for paragraph in paragraphs:
        # If this paragraph starts and ends with a code block, add it as is.
        if paragraph.startswith('```') and paragraph.endswith('```'):
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the beginning of a code block add it as is.
        if paragraph.startswith('```'):
            in_code_block = True
            wrapped_paragraphs.append(paragraph)
            continue
        # If this is the end of a code block stop skipping text.
        if paragraph.endswith('```'):
            in_code_block = False
            wrapped_paragraphs.append(paragraph)
            continue
        # If we are currently in a code block add the paragraph as is.
        if in_code_block:
            wrapped_paragraphs.append(paragraph)
        else:
            # Otherwise, apply text wrapping to the paragraph.
            wrapped_paragraph = textwrap.fill(paragraph)
            wrapped_paragraphs.append(wrapped_paragraph)
    # Join all paragraphs into a single string
    wrapped_text = '\n\n'.join(wrapped_paragraphs)
    return wrapped_text

def read_lines_width() -> int:
    return 10

def read_lines(file_path: str, start_line: int, end_line: int) -> str:
    """
    Read lines from a file and return a string containing the lines between start_line and end_line.

    Args:
        file_path (str): The path of the file to read.
        start_line (int): The line number of the first line to include (1-indexed).
        end_line (int): The line number of the last line to include.

    Returns:
        str: A string containing the lines between start_line and end_line.

    """
    # open the file for reading
    with open(file_path, 'r') as f:
        # read all the lines from the file
        lines = f.readlines()
        # remove trailing newline characters
        lines = [line.rstrip() for line in lines]
        # add line numbers
        lines = [f"   {index+1:<6} {line}" for index, line in enumerate(lines)]
    # convert start_line to 0-based indexing
    start_line = max(0, start_line - 1)
    # ensure end_line is within range
    end_line = min(len(lines), end_line)
    # return the requested lines as a string
    return '\n'.join(lines[start_line:end_line])

async def explain(source_code: str, traceback: str, exception: str, really_run = True) -> None:
    import httpx
    user_prompt = "Explain what the root cause of this error is, given the following source code context for each stack frame and a traceback, and propose a fix. In your response, never refer to the frames given below (as in, 'frame 0'). Instead, always refer only to specific lines and filenames of source code.\n"
    user_prompt += '\n'
    user_prompt += 'Source code for each stack frame:\n```\n'
    user_prompt += source_code + '\n```\n'
    user_prompt += traceback + '\n\n'
    user_prompt += 'stop reason = ' + exception + '\n'
    text = ''
    
    if not really_run:
        print(user_prompt)
        return

    if not 'OPENAI_API_KEY' in os.environ:
        print('You need a valid OpenAI key to use ChatDBG. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        return

    model = get_model()
    if not model:
        return
       
    try:
        completion = await openai_async.chat_complete(openai.api_key, timeout=30, payload={'model': f'{model}', 'messages': [{'role': 'user', 'content': user_prompt}]})
        json_payload = completion.json()
        text = json_payload['choices'][0]['message']['content']
    except (openai.error.AuthenticationError, httpx.LocalProtocolError, KeyError):
        # Something went wrong.
        print()
        print('You need a valid OpenAI key to use ChatDBG. You can get a key here: https://openai.com/api/')
        print('Set the environment variable OPENAI_API_KEY to your key value.')
        import sys
        sys.exit(1)
    except Exception as e:
        print(f'EXCEPTION {e}, {type(e)}')
        pass
    print(word_wrap_except_code_blocks(text))

